import seqtools, sys, getopt

aminoacids = 'ARNDCEQGHILKMFPSTWYV'
dnafirst = 'GACTRNDEQHILKMFPSWYV'

def merge(bins, indexA, indexB, statedicts):

  newPenalty = bins[indexA].score + bins[indexA].penalty

  for b in bins:
    b.removeCombos(bins[indexA].name)
    b.removeCombos(bins[indexB].name)

  newname = sortstr(bins[indexA].name + bins[indexB].name)

  del bins[max(indexA, indexB)]
  del bins[min(indexA, indexB)]
  
  newbin = bin(newname)
  
  for b in bins:
    comboName = sortstr(b.name + newname)

    # uncomment to include sites with fewer characters
    #comboScore = getClassScore(comboName, statedicts) - newPenalty - b.penalty
    try:
      comboScore = statedicts[len(comboName) - 1][comboName]
      newbin.addCombo(comboName, comboScore)
      b.addCombo(comboName, comboScore)
      b.refreshByFreq()
    except KeyError:
      pass
  
  newbin.refreshByFreq()
  newbin.setPenalty(newPenalty)

  bins.append(newbin)
  bins.sort()
  bins.reverse()
  
  #print 'new bin:', newbin.name, 'score:', newbin.score
  #print newbin.byFreq

  return newname

def getClassScore(classStates, statedicts):

  score = 0

  for nstates in xrange(1, len(classStates)):
    classDict = statedicts[nstates]
    for className in classDict:
      inclass = True
      for char in className:
        if char not in classStates:
          inclass = False
          break
      if inclass:
        score += classDict[className]

  return score

def sortstr(string):
  chararray = list(string)
  chararray.sort()
  
  return ''.join(chararray)

def buildbins(comboDict):
  binsDict = {}
  for char in aminoacids:
    binsDict[char] = bin(char)

  for combo in comboDict:
    for state in combo:
      binsDict[state].addCombo(combo, comboDict[combo])
      
  bins = []
  for char in aminoacids:
    bins.append(binsDict[char])
    bins[-1].refreshByFreq()
    
  return bins

def transpose(alignment):
  trans = []

  for i in xrange(len(alignment)):
    try:
      trans.append(alignment[i])
  
    except:
      for seq in alignment.seqlist:
        print len(seq)
        
      raise
    
  return trans
  
def getuniquechars(listObj):
  unique = []
  for item in listObj:
    if item not in unique and item in aminoacids:
      unique.append(item)
      
  return unique
  
def getsignificantchars(listObj):
  uniquechars = getuniquechars(listObj)
  counts = [(listObj.count(char), char) for char in uniquechars]
  counts.sort()
  counts.reverse()
  
  totalfreq = 0.0
  significantchars = []
  nchar = 0
  for char in listObj:
    if char in aminoacids:
      nchar += 1
  nchar = float(nchar)
  
  countptr = 0
  while totalfreq < 0.9:
    try:
      significantchars.append(counts[countptr][1])
    except:
      print 'significantchars:', significantchars
      print 'countptr:', countptr
      print 'counts:', counts
      print 'totalfreq:', totalfreq
      raise
    totalfreq += counts[countptr][0]/nchar
    countptr += 1

  return significantchars

def buildstatedicts(columns):
  statedicts = [{} for i in xrange(20)]
  for col in columns:
    colchars = getsignificantchars(col)
    colchars.sort()
    nstate = len(colchars)
    
    try:
      statedicts[nstate - 1][''.join(colchars)] += 1
    except KeyError:
      statedicts[nstate - 1][''.join(colchars)] = 1
      
  return statedicts
  

def substrings(string):
  for substring in substringsRec(string, 3):
    yield substring
        
        
def substringsRec(string, nchar):
  #print 'in substringsRec, string = %s, nchar = %d.'  % (string, nchar)
  if nchar == 1:
    for i in xrange(len(string)):
      yield string[i]
      
  else:
    for i in xrange(len(string) - nchar + 1):
      for remainder in substringsRec(string[i+1:], nchar - 1):
        yield string[i] + remainder

 
class bin:
  def __init__(self, name):
    self.name = name
    self.byFreq = []
    self.byName = {}
    self.score = 0
    self.penalty = 0
    
  def setScore(self):
    try:
      self.score = self.byFreq[0][0]
    except IndexError:
      self.score = 0
  
  def refreshByName(self):
    self.byName = {}
    for combination in self.byFreq:
      self.byName[combination[1]] = combination[0]
      
  def refreshByFreq(self):
    self.byFreq = []
    for name in self.byName:
      self.byFreq.append((self.byName[name], name))
      
    self.byFreq.sort()
    self.byFreq.reverse()
    self.setScore()

  def setPenalty(self, penalty):
    self.penalty = penalty
    
  def getComboScore(self, combo):
    return self.byName[combo]
    
  def addCombo(self, combo, score):
    """Note that byFreq isn't updated here!"""
    self.byName[combo] = score
    
  def removeCombos(self, comboItem):
    """Delete entries including all letters in comboItem."""
    # could probably be ANY letters, but...
  
    for i in xrange(len(self.byFreq) - 1, -1, -1):
      for letter in comboItem:
        if letter in self.byFreq[i][1]:
          # this will always happen with the first letter
          del self.byFreq[i]
          break
        
    self.refreshByName()
    
  # these methods for sorting purposes
  def __lt__(self, other):
    return self.score < other.score
    
  def __le__(self, other):
    return self.score <= other.score
  
  def __eq__(self, other):
    return self.score == other.score
  
  def __ne__(self, other):
    return self.score != other.score
  
  def __gt__(self, other):
    return self.score > other.score
  
  def __ge__(self, other):
    return self.score >= other.score

def printbins(bins):
  binNames = [b.name for b in bins]
  binNames.sort()

  for name in binNames:
    sys.stdout.write(' %s' % name)
  sys.stdout.write('\n')
  
def recode(align, bins):
  binNames = [b.name for b in bins]
  binNames.sort()
  
  recodedict = {}
  
  for i in xrange(len(binNames)):
    for char in binNames[i]:
      recodedict[char] = dnafirst[i]
        
  for seq in align.seqlist:
    recodedseq = []
    
    for char in seq.sequence:
      try:
        recodedseq.append(recodedict[char])
      except KeyError:
        recodedseq.append(char)
        
    seq.sequence = ''.join(recodedseq)
  

def dogetopt(argv, reqdopts, legalopts):
  """Return a dictionary of options, check to make sure reqdopts are there."""
  
  if '-h' in argv:
    usage()
  
  opts, args = getopt.getopt(argv, legalopts)
  
  optdict = {}
  
  for (opt, val) in opts:
    optdict[opt[1:]] = val
   
  for opt in reqdopts:
    if opt not in optdict:
      raise getopt.GetoptError('Missing required option!')
      
  return (optdict, args)
  
def usage():
  sys.stderr.write('Usage: python QuASAR -i alignment [-o outfile] [-z alphabetsize]\n\n')
  sys.stderr.write('Optional arguments:\n')
  sys.stderr.write('  -o outfile      File name for recoded alignment [default: do not write file]\n')
  sys.stderr.write('  -z alphabetsize Number of characters in the reduced alphabet [default: 1]\n\n')
  
  sys.stderr.write('  Alphabetsize must be greater than 1 and less than 20. If alphabetsize=4,\n')
  sys.stderr.write('  nucleotide characters will be used. If either outfile or alphabet size is\n')
  sys.stderr.write('  not specified, the alignment will not be recoded. Rather, the recoding\n')
  sys.stderr.write('  scheme that would be used will just be printed to the terminal.\n')
  
  sys.exit(1)


if __name__ == '__main__':
  try:
    optdict, args = dogetopt(sys.argv[1:], 'i', 'i:o:z:h')
  except getopt.GetoptError:
    usage()

  if 'h' in optdict:
    usage()

  if 'o' in optdict:
    outfile = optdict['o']
  else:
    outfile = None
    
  # size of the reduced alphabet
  if 'z' in optdict:
    try:
      nbins = int(optdict['z'])
      if nbins > 19 or nbins < 2:
        usage()
    except ValueError:
      usage()
  else:
    nbins = 1
    
  align = seqtools.getseqs(optdict['i'])
  cols = transpose(align)
  statedicts = buildstatedicts(cols)
  bins = buildbins(statedicts[1]) # consider only 2-state characters
  
  bins.sort()
  bins.reverse()


  while len(bins) > nbins: 
    mstr = 'Merging %s and %s... (score = %d)' % (bins[0].name, bins[1].name, bins[0].score)
    sys.stdout.write('%-60s' % mstr)
    merge(bins, 0, 1, statedicts)
    
    printbins(bins)
  
  
  if nbins > 1 and outfile:
     recode(align, bins)
     seqtools.writephy(align, outfile)
  sys.exit(0)
  

  for b in bins:
    print 'bin name:', b.name, 'score:', b.score
    print b.byFreq
 
  lastMergedScore = bins[0].score
  lastMerged = merge(bins, 0, 1, statedicts)
  print '================= Merged %s =======================' % lastMerged

  for b in bins:
    print 'bin name:', b.name, 'score:', b.score
    print b.byFreq

  lastMergedScore = bins[4].score
  lastMerged = merge(bins, 4,5, statedicts)
  print '================= Merged %s =======================' % lastMerged
  
  for b in bins:
    print 'bin name:', b.name, 'score:', b.score
    print b.byFreq
 
 # get a number of bins from the user; if not specified, just print the binning scheme
 # if specified, loop only until there are that many bins
 

