import getopt, glob, os, sys

from seqtools import *

def alignseqs(genename, aligner):
  # infile: name of fasta-format file to be aligned
  # don't forget to convert output to pir
  # careful with hmmer format! funny added x characters... convert to - ?
  
  if aligner == 'muscle' or aligner == 'mhmmer':
    runmuscle(genename)
    print '    Finished MUSCLE alignment.'
    hmmerin = '%s-muscle.pir' % genename
    sys.stdout.flush()
  
  elif aligner == 'probcons' or aligner == 'phmmer':
    runprobcons(genename)
    print '    Finished ProbCons alignment.'
    hmmerin = '%s-probcons.pir' % genename
    sys.stdout.flush()
  
  elif aligner == 'clustal' or aligner == 'chmmer':
    runclustal(genename)
    print '    Finished ClustalW alignment.'
    hmmerin = '%s-clustal.pir' % genename
    sys.stdout.flush()
  
  elif aligner == 'tcoffee' or aligner == 'thmmer':
    runtcoffee(genename)
    print '    Finished TCOFFEE alignment.'
    hmmerin = '%s-tcoffee.pir' % genename
    sys.stdout.flush()
  
  else:
    sys.stderr.write('Invalid alignment program: %s\n' % aligner)
    sys.exit(1)
   
  if aligner[1:] == 'hmmer':
    runhmmer(genename, hmmerin)
    print '    Finished refining alignment with HMMER.'

    
  #else:
  #  os.remove(hmmerin)
  
def runbin(binary, cmds=[], debug=False):
  
  instream, outstream = os.popen4(binary)
  
  for cmd in cmds:
    instream.write('%s\n' % cmd)
    
  instream.close()
  
  for line in outstream:
    if debug:
      sys.stderr.write(line)
  
  outstream.close()
  
  try:
    os.wait()
  except OSError:
    pass

def runclustal(genename):
  clustalcmd = 'clustalw -infile=%s.fas -outfile=%s-clustal.aln' % \
             (genename, genename)
  
  while True:
    try:
      runbin(clustalcmd)
  
      # make PIR copy
      writefasta(getseqs('%s-clustal.aln' % genename), '%s-clustal.pir' % genename,\
                 variation='pir')
      #os.remove('%s-clustal.pir' % genename) # delete clustal file, clustal sucks!      
      
      break
    
    # why IndexError?  Is this right?
    except (IndexError, IOError): # clustal crashes occasionally, can't make PIR from ALN
      print '    (problem with ClustalW, trying again)'
      continue
      
  
  # clean up
  os.remove('%s.dnd' % genename)
  os.remove('%s-clustal.aln' % genename)

def rungblocks(genename, aligner):
  align = getseqs('%s.fas'  % genename)
  if aligner[1:] == 'hmmer':
    infile = '%s-hmmer.pir' % genename
  else:
    infile = '%s-%s.pir' % (genename, aligner)
    
  ntaxa = len(align.taxlist)
  
  cmds = ['b', '5', '5', '2', '%d' % (ntaxa / 2 + 1), 'g', 'm', 'q']
  #runbin('Gblocks %s-hmmer.pir' % genename, cmds) 
  #os.rename('%s-hmmer.pir-gb' % genename, '%s-gb.pir' % genename)
  #os.remove('%s-hmmer.pir-gb.htm' % genename)
  runbin('Gblocks %s' % infile, cmds)
  
  os.rename('%s-gb' % infile, '%s-gb.pir' % genename)
  os.remove('%s-gb.htm' % infile)
  
  return getseqs('%s-gb.pir' % genename)
  
def runhmmer(genename, clustalalign):

  iternum = 0
  oldalign = clustalalign
  
  while iternum < 50: # stop after 50 iterations, no matter what
    iternum += 1
    align = '%s-hmmer%d.align' % (genename, iternum)
    model = '%s-hmmer%d.model' % (genename, iternum)
    
    # build model - always with previous alignment (with clustal for first iter)
    runbin('hmmbuild --amino %s %s' % (model, oldalign))
    
    # build alignment
    runbin('hmmalign --outformat SELEX -o %s %s %s' % (align, model, clustalalign))
    
    if iternum == 1:
      oldalign = align
      continue
    
    # diff exits with code of 0 iff files are identical
    exitcode = os.system('diff %s %s > /dev/null' % (align, oldalign))    
    if exitcode == 0:
      break # when iterations converge (no more change in alignments)
    
    # if we don't break
    oldalign = align
    
  # convert to PIR format
  writefasta(getseqs(align), '%s-hmmer.pir' % genename, variation='pir')
  
  # clean up
  #os.remove('%s-clustal.aln' % genename)
  os.remove(clustalalign)
  
  for f in glob.glob('%s-hmmer*.align' % genename):
    os.remove(f)
    
  for f in glob.glob('%s-hmmer*.model' % genename):
    os.remove(f)
    
def runmuscle(genename):
  while True:
    try:
      runbin('muscle -in %s.fas -out %s-muscle.fa' % (genename, genename))

      musclealign = getseqs('%s-muscle.fa' % genename)
      writefasta(musclealign, '%s-muscle.pir' % genename, \
                 variation='pir')
      # write clustal-format alignment to start hmmer with.
      #writeclustal(musclealign, '%s-muscle.aln' % genename)
      break
    except IOError:
      continue
  
  os.remove('%s-muscle.fa' % genename)
  
def runprobcons(genename):
  while True:
    try:
      runbin('probcons %s.fas > %s-probcons.fa' % (genename, genename))
      align = getseqs('%s-probcons.fa' % genename)
      for seq in align.seqlist:
        #print 'len(seq):', len(seq), 'len(align):', len(align)
        if len(seq) < len(align):
          seq.pad(len(align))
          sys.stderr.write('Sequence %s too short! Padding.\n' % seq.name)
      writefasta(align, '%s-probcons.pir' \
                 % genename, variation='pir')
      break
    # if probcons fails, IndexError when trying to read alignment from .fa file
    except (IOError, IndexError):
      continue
    
  os.remove('%s-probcons.fa' % genename)
  
def runtcoffee(genename):
  for i in xrange(100): # tcoffee is buggy... retry until it works
    try:
      runbin('t_coffee -output=pir -in fast_pair %s.fas' % genename)
  
      #writeclustal(getseqs('%s.pir' % genename), '%s-tcoffee.aln' % genename)
  
      os.rename('%s.pir' % genename, '%s-tcoffee.pir' % genename)
      os.remove('%s.dnd' % genename)
      break

    except OSError:
      print '    (Attempt %d: problem with TCOFFEE, trying again...)' % (i + 1)
      continue

def cleanalignment(alignment, minchars):
  # remove mostly-gap positions, remove soap positions?
  # alignment: an Alignment object
  
  taxa = [seq.name for seq in alignment.seqlist]
  alignT = transpose(alignment)
  
  for i in xrange(len(alignT) - 1, -1, -1): # indices of columns
    col = alignT[i]
    charcount = 0
    for char in col:
      if char != 'X' and char != '-':
        charcount += 1
    
    # delete columns that have too few real characters
    if charcount < minchars:
      del alignT[i] 
    
  newseqs = [[] for i in xrange(len(taxa))]
  
  for col in alignT: # actually rows in the transpose, alignT
    for i in xrange(len(col)):
      newseqs[i].append(col[i])
  
  
  return Alignment([Sequence(taxa[i], ''.join(newseqs[i])) for i in xrange(len(newseqs))])

def transpose(alignment):
  # alignment: an Alignment object
  # returns a list of columns in the alignment
  # a column is a list of characters found in that column
  
  columns = [[] for i in xrange(alignment.length)]
  for seq in alignment.seqlist:
    for i in range(len(seq.sequence)):
      columns[i].append(seq.sequence[i])
      
  return columns

def dogetopt(argv, reqdopts, legalopts):
  """Return a dictionary of options, check to make sure reqdopts are there."""
  
  if '-h' in argv:
    usage()
  
  opts, args = getopt.getopt(argv, legalopts)
  
  optdict = {}
  
  for (opt, val) in opts:
    optdict[opt[1:]] = val
   
  for opt in reqdopts:
    if opt not in optdict:
      raise getopt.GetoptError('Missing required option!')
      
  return (optdict, args)
  
def usage():
  sys.stderr.write('Usage: python alignmonkey.py -i ext -a [application]\n\n')
  sys.stderr.write('  -i ext  Extension of input multiple sequence files [default: fas]\n')
  sys.stderr.write('  -a app  Program to use to align sequences\n\n')
  sys.stderr.write('  Available applications (assuming you have them installed):\n')
  sys.stderr.write('    muscle\n')
  sys.stderr.write('    mhmmer (Muscle + HMMER) [default]\n')
  sys.stderr.write('    probcons\n')
  sys.stderr.write('    phmmer (Probcons + HMMER)\n')
  sys.stderr.write('    clustal (ClustalW)\n')
  sys.stderr.write('    chmmer (ClustalW + HMMER)\n')
  sys.stderr.write('    tcoffee\n')
  sys.stderr.write('    thmmer (TCOFFEE + HMMER)\n')

  sys.exit(1)
  
  
if __name__ == '__main__':

  try:
    optdict, args = dogetopt(sys.argv[1:], '', 'i:a:h')
  except getopt.GetoptError:
    usage()
    
  if 'h' in optdict:
    usage()
    
  if 'i' in optdict:
    alignext = optdict['i']
  else:
    alignext = 'fas'
    
  validaligners = ('muscle', 'mhmmer', 'probcons', 'phmmer', 'clustal', \
                   'chmmer', 'tcoffee', 'thmmer')
  
  if 'a' in optdict:
    alignprog = optdict['a']
    if alignprog not in validaligners:
      usage()
  else:
    alignprog = 'mhmmer'

  
  try:
    os.mkdir('alignments')
  except OSError:
    pass

  genelist = [f[:-4] for f in glob.glob('*.%s' % alignext)]

  if not genelist:
    usage()

  try:
    os.mkdir('library')
    # remove gaps and copy FASTA files to library directory
    for f in genelist:
      align = getseqs('%s.%s' % (f, alignext))
      for seq in align.seqlist:
        seq.sequence = seq.sequence.replace('-', '')
      writefasta(align, 'library/%s.fas' % f)
  except OSError:
    pass
  
  minchars = 4 # minimum number of characters in an alignment column

  os.chdir('library') # move to directory with unaligned sequences
    
  cleanaligns = []
  
  #print 'genelist:', genelist
  sys.stdout.flush()
  
  for gene in genelist:
    print 'Aligning %s...' % gene
    alignseqs(gene, alignprog)
    print '    Running Gblocks & cleaning up...',
    cleanaligns.append(cleanalignment(rungblocks(gene, alignprog), minchars)) 
    writephy(cleanaligns[-1], '../alignments/%s-clean.seq' % gene)
    print 'done.'
    
    for f in glob.glob('%s*.pir' % gene):
      os.rename(f, '../alignments/%s' % f)
  
  print 'Concatenating alignments...',
  os.chdir('../alignments')
  writephy(catseq(cleanaligns), 'aligncat.seq')
  print 'done.'
  

  
  
